"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.arithmeticExternals = exports.symb = void 0;
const Imports_1 = require("../Compiler/Passes/Imports");
const Utils_1 = require("../Compiler/Utils");
const JSTranslator_1 = require("../Translator/JSTranslator");
exports.symb = (f) => ({ name: f, args: [] });
const arithbinop = (t, op) => {
    const [a, b] = t.args;
    if (Utils_1.isFun(a) && Utils_1.isFun(b)) {
        try {
            const an = BigInt(a.name);
            const bn = BigInt(b.name);
            const res = op(an, bn);
            return exports.symb(`${res}`);
        }
        catch (e) {
            return t;
        }
    }
    return t;
};
const boolbinop = (t, op) => {
    const [a, b] = t.args;
    if (Utils_1.isFun(a) && Utils_1.isFun(b)) {
        const an = BigInt(a.name);
        const bn = BigInt(b.name);
        const res = op(an, bn);
        return res ? Imports_1.True : Imports_1.False;
    }
    return t;
};
const nativeArithmeticExternals = {
    'add': t => arithbinop(t, (a, b) => a + b),
    'sub': t => arithbinop(t, (a, b) => a - b),
    'mult': t => arithbinop(t, (a, b) => a * b),
    'div': t => arithbinop(t, (a, b) => a / b),
    'mod': t => arithbinop(t, (a, b) => a % b),
    'pow': t => arithbinop(t, (a, b) => a ** b),
    'gtr': t => boolbinop(t, (a, b) => a > b),
    'geq': t => boolbinop(t, (a, b) => a >= b),
    'lss': t => boolbinop(t, (a, b) => a < b),
    'leq': t => boolbinop(t, (a, b) => a <= b)
};
const jsarithbinop = (op) => {
    return (name) => (`function ${name}(a, b) {
            if (isNat(a) && isNat(b)) {
                return ${op === '/' ? 'typeof a === "number" ? Math.floor(a / b) : a / b' : `a ${op} b`};
            }
        
            return { name: "${name}", args: [a, b] };
        }`);
};
const jsboolbinop = (op) => {
    return (name) => (`function ${name}(a, b) {
            if (isNat(a) && isNat(b)) {
                return (a ${op} b) ? ${JSTranslator_1.nullaryVarName('True')} : ${JSTranslator_1.nullaryVarName('False')};
            }
        
            return { name: "${name}", args: [a, b] };
        }`);
};
const jsArithmeticExternals = {
    'add': jsarithbinop('+'),
    'sub': jsarithbinop('-'),
    'mult': jsarithbinop('*'),
    'div': jsarithbinop('/'),
    'mod': jsarithbinop('%'),
    'pow': jsarithbinop('**'),
    'gtr': jsboolbinop('>'),
    'geq': jsboolbinop('>='),
    'lss': jsboolbinop('<'),
    'leq': jsboolbinop('<=')
};
const haskellArithmeticExternals = {
    "sub": name => `${name} (Fun a []) (Fun b []) =
          Fun (show ((read a :: Int) - (read b :: Int))) []`,
    "add": name => `${name} (Fun a []) (Fun b []) =
          Fun (show ((read a :: Int) + (read b :: Int))) []`,
    "mult": name => `${name} (Fun a []) (Fun b []) =
          Fun (show ((read a :: Int) * (read b :: Int))) []`,
    "mod": name => `${name} (Fun a []) (Fun b []) =
          Fun (show ((read a :: Int) \`mod\` (read b :: Int))) []`,
    "div": name => `${name} (Fun a []) (Fun b []) =
          Fun (show ((read a :: Int) \`div\` (read b :: Int))) []`,
    "pow": name => `${name} (Fun a []) (Fun b []) =
          Fun (show ((read a :: Int) ^ (read b :: Int))) []`,
    "lss": name => `${name} (Fun a []) (Fun b []) =
          Fun (show ((read a :: Int) < (read b :: Int))) []`,
    "leq": name => `${name} (Fun a []) (Fun b []) =
          Fun (show ((read a :: Int) <= (read b :: Int))) []`,
    "gtr": name => `${name} (Fun a []) (Fun b []) =
          Fun (show ((read a :: Int) > (read b :: Int))) []`,
    "geq": name => `${name} (Fun a []) (Fun b []) =
          Fun (show ((read a :: Int) >= (read b :: Int))) []`,
};
const ocamlIntBinop = (op, reversed = false) => {
    return (name) => 'let ' + name + ' (a, b) = match (a, b) with\n' +
        '   | ((Fun (a, [])), (Fun (b, []))) ->\n' +
        (reversed ?
            '       Fun ((string_of_int ((' + op + ' (int_of_string a)) (int_of_string b))), [])' :
            '       Fun ((string_of_int ((int_of_string a) ' + op + ' (int_of_string b))), [])') + '\n' +
        '   | _ -> Fun ("@' + name + '", []);;';
};
const ocamlBoolOf = (expr) => `(if ${expr} then "True" else "False")`;
const ocamlBoolBinop = (op) => {
    return (name) => 'let ' + name + ' (a, b) = match (a, b) with\n' +
        '   | ((Fun (a, [])), (Fun (b, []))) ->\n' +
        '       Fun (' + ocamlBoolOf('((int_of_string a) ' + op + ' (int_of_string b))') + ', [])\n' +
        '   | _ -> Fun ("@' + name + '", []);;';
};
const ocamlArithmeticExternals = {
    "sub": ocamlIntBinop('-'),
    "add": ocamlIntBinop('+'),
    "mult": ocamlIntBinop('*'),
    "mod": ocamlIntBinop('mod'),
    "div": ocamlIntBinop('/'),
    "pow": name => `
let rec pow a = function
    | 0 -> 1
    | 1 -> a
    | n -> 
      let b = pow a (n / 2) in
      b * b * (if n mod 2 = 0 then 1 else a);;
    
      ${ocamlIntBinop('pow', true)(name)}
`.trim().trimEnd(),
    "lss": ocamlBoolBinop('<'),
    "leq": ocamlBoolBinop('<='),
    "gtr": ocamlBoolBinop('>'),
    "geq": ocamlBoolBinop('>=')
};
exports.arithmeticExternals = target => {
    return {
        'native': nativeArithmeticExternals,
        'js': jsArithmeticExternals,
        'ocaml': ocamlArithmeticExternals,
        'haskell': haskellArithmeticExternals
    }[target];
};
